from motor_odm.helpers import inherit_class


def test_mongo_inherit():
    class Base:
        parent_field = "Parent"
        field = "Test"

    class Child:
        child_field = "Child"
        field = "Test 2"

    New = inherit_class("Meta", Child, Base)

    assert New.__name__ == "Meta"
    assert New.child_field == "Child"
    assert New.parent_field == "Parent"
    assert New.field == "Test 2"



# -- Path setup --------------------------------------------------------------

import os
import sys
from datetime import datetime

from pkg_resources import get_distribution

sys.path.insert(0, os.path.abspath(".."))

# -- Project information -----------------------------------------------------

project = "Motor-ODM"
copyright = f"{datetime.now().year}, Kim Wittenburg"
author = "Kim Wittenburg"

release = get_distribution("motor_odm").version
version = ".".join(release.split(".")[:2])

# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosummary",
    "sphinxcontrib.apidoc",
]

templates_path = ["templates"]

exclude_patterns = ["build", "Thumbs.db", ".DS_Store"]

warning_is_error = True

# -- Autodoc Options ---------------------------------------------------------

autodoc_member_order = "groupwise"

# -- APIDoc Options ----------------------------------------------------------

apidoc_module_dir = "../motor_odm"
apidoc_output_dir = "reference"
apidoc_separate_modules = True
apidoc_toc_file = False
apidoc_module_first = True
apidoc_extra_args = ["--templatedir", "templates"]

os.environ["SPHINX_APIDOC_OPTIONS"] = ",".join(
    ["members", "undoc-members", "show-inheritance"]
)
# -- Options for HTML output -------------------------------------------------

html_theme = "sphinx_rtd_theme"
extensions.append("sphinx_rtd_theme")

html_theme_options = {"collapse_navigation": False, "navigation_depth": 5}

# html_static_path = ["static"]

intersphinx_mapping = {
    "pymongo": ("https://pymongo.readthedocs.io/en/stable/", None),
    "py": ("https://docs.python.org/3/", None),
}



import json
import os

import pytest
from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorDatabase

from motor_odm import Document


@pytest.fixture(scope="function")
async def db() -> AsyncIOMotorDatabase:
    client = AsyncIOMotorClient("localhost")
    db = client["test"]
    await client.drop_database(db)
    Document.use(db)
    yield db
    await client.drop_database(db)


@pytest.fixture(scope="function")
async def data(db: AsyncIOMotorDatabase) -> None:
    with open(os.path.join(os.path.dirname(__file__), "data.json")) as f:
        file_data = json.load(f)
    for key, value in file_data.items():
        await db[key].insert_many(value)



"""
This module contains various supporting functions that can be used independently of the Motor-ODM framework. Some of
these utilities can be found in similar form in other packages or frameworks and are adapted here to reduce the number
of dependencies.
"""
import inspect
from inspect import isclass, ismodule
from types import ModuleType
from typing import TYPE_CHECKING, Any, Callable, Optional, Tuple, TypeVar, Union

if TYPE_CHECKING:
    T = TypeVar("T", bound=type)
    C = TypeVar("C", bound=Callable)
    D = TypeVar("D", bound=Union[object, Callable[[Any], Any]])


def inherit_class(name: str, self: Optional["T"], *parents: "T") -> "T":
    """
    Performs a pseudo-inheritance by creating a new class that inherits from ``self`` and ``parents``. This is useful to
    support intuitive inheritance on inner classes (typically named ``Meta``).

    Note that this method neither returns ``self`` nor any of the ``parents`` but a new type that inherits from both.

    :param name: The name of the newly created type.
    :param self: The primary base class (fields in this class take preference over the ``parents``' fields.
    :param parents: The secondary base classes. Field preferences are determined by the order of the parent classes.
    :return: A new type inheriting from ``self`` and ``parents``.
    """
    base_classes: Tuple["T", ...]
    if not self:
        base_classes = parents
    elif self == parents[0]:
        base_classes = (self, *parents[1:])
    else:
        base_classes = self, *parents
    return type(name, base_classes, {})  # type: ignore


def monkey_patch(
    cls: Union[type, ModuleType], name: Optional[str] = None
) -> Callable[["C"], "C"]:
    """
    Monkey patches class or module by adding to it decorated function. Anything overwritten can be accessed via a
    ``.original`` attribute of the decorated object.

    :param cls: The class or module to be patched.
    :param name: The name of the attribute to be patched.
    :return: A decorator that monkey patches ``cls.name`` and returns the decorated function.
    """
    assert isclass(cls) or ismodule(
        cls
    ), "Attempting to monkey patch non-class and non-module"

    def decorator(value: "D") -> "D":
        assert inspect.isfunction(value)

        func = getattr(value, "fget", value)  # Support properties
        if name:
            func_name = name
        elif func.__name__.startswith(f"{cls.__name__}__"):
            func_name = func.__name__[len(f"{cls.__name__}__") :]
        else:
            func_name = func.__name__

        func.__name__ = func_name
        func.original = getattr(cls, func_name, None)

        setattr(cls, func_name, value)
        return value

    return decorator



import pytest
from motor.motor_asyncio import AsyncIOMotorDatabase

pytestmark = pytest.mark.asyncio


@pytest.mark.usefixtures("data")
async def test_get_by_name(db: AsyncIOMotorDatabase):
    result = await db["test"].find_one({"name": "Test 1"})
    assert result is not None
    assert isinstance(result, dict)



"""
This module contains the base class for interacting with Motor-ODM: :class:`Document`. The :class:`Document` class is
the main entry point to Motor-ODM and provides its main interface.
"""

from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Callable,
    Iterator,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from bson import CodecOptions, ObjectId
from motor.core import AgnosticCollection, AgnosticDatabase
from pydantic import BaseModel, Field
from pydantic.main import ModelMetaclass
from pymongo import ReadPreference, WriteConcern
from pymongo.read_concern import ReadConcern
from pymongo.results import InsertManyResult

from motor_odm.helpers import monkey_patch

from .helpers import inherit_class
from .query import create_query

if TYPE_CHECKING:
    from pydantic.typing import (  # noqa: F401
        DictStrAny,
        AbstractSetIntStr,
        DictIntStrAny,
    )
    from motor_odm.query import Query

    GenericDocument = TypeVar("GenericDocument", bound="Document")
    MongoType = Type["MongoBase"]

__all__ = ["DocumentMetaclass", "Document"]


@monkey_patch(ObjectId)
def __get_validators__() -> Iterator[Callable[[Any], ObjectId]]:
    """
    Returns a generator yielding the validators for the `ObjectId` type.
    """

    def validate(value: Union[str, bytes]) -> ObjectId:
        """
        Creates an `ObjectId` from a value. If the
        :param value: An `ObjectId` or a string.
        :return: An `ObjectId`.
        :raises TypeError: If `value` os not a string type.
        :raises InvalidId: If `value` does not represent a valid `ObjectId`.
        """

        return ObjectId(value)

    yield validate


class MongoBase:
    """This class defines the defaults for collection configurations.

    Each collection (defined by a subclass of :class:`Document`) can override these using an inner class named
    ``Mongo``. Attributes are implicitly and transitively inherited from the Mongo classes of base classes.
    """

    collection: Optional[str]
    """The name of the collection for a document. This attribute is required."""

    codec_options: Optional[CodecOptions] = None
    """The codec options to use when accessing the collection. Defaults to the database's :attr:`codec_options`."""

    read_preference: Optional[ReadPreference] = None
    """The read preference to use when accessing the collection. Defaults to the database's :attr:`read_preference`."""

    write_concern: Optional[WriteConcern] = None
    """The write concern to use when accessing the collection. Defaults to the database's :attr:`write_concern`."""

    read_concern: Optional[ReadConcern] = None
    """The read concern to use when accessing the collection. Defaults to the database's :attr:`read_concern`."""


class DocumentMetaclass(ModelMetaclass):
    """The meta class for :class:`Document`. Ensures that the ``Mongo`` class is automatically inherited."""

    def __new__(
        mcs, name: str, bases: Sequence[type], namespace: "DictStrAny", **kwargs: Any
    ) -> "DocumentMetaclass":
        mongo: MongoType = MongoBase
        for base in reversed(bases):
            if base != BaseModel and base != Document and issubclass(base, Document):
                # noinspection PyTypeChecker
                mongo = inherit_class("Mongo", base.__mongo__, mongo)
        # noinspection PyTypeChecker
        mongo = inherit_class("Mongo", namespace.get("Mongo"), mongo)

        if (namespace.get("__module__"), namespace.get("__qualname__")) != (
            "motor_odm.document",
            "Document",
        ):
            if not hasattr(mongo, "collection"):
                raise TypeError(f"{name} does not define a collection.")

        return super().__new__(mcs, name, bases, {"__mongo__": mongo, **namespace}, **kwargs)  # type: ignore


class Document(BaseModel, metaclass=DocumentMetaclass):
    """This is the base class for all documents defined using Motor-ODM.

    A :class:`Document` is a pydantic model that can be inserted into a MongoDB collection. This class provides an easy
    interface for interacting with the database. Each document has an :attr:`Document.id` (named ``_id`` in MongoDB) by default by which it can be uniquely
    identified in the database. The name of this field cannot be customized however you can override it if you don't
    want to use :class:`ObjectID <bson.objectid.ObjectId>` values for your IDs.
    """

    class Config:
        """:meta private:"""

        validate_all = True
        validate_assignment = True
        allow_population_by_field_name = True

    if TYPE_CHECKING:
        # populated by the metaclass, defined here to help IDEs only
        __mongo__: MongoType

    __db__: Optional[AgnosticDatabase]
    __collection__: Optional[AgnosticCollection] = None

    id: ObjectId = Field(None, alias="_id")
    """The document's ID in the database.

    By default this field is of type :class:`ObjectId <bson.objectid.ObjectId>` but it can be overridden to supply your
    own ID types. Note that if you intend to override this field you **must** set its alias to ``_id`` in order for your
    IDs to be recognized as such by MongoDB.
    """

    @classmethod
    def use(cls: Type["Document"], db: AgnosticDatabase) -> None:
        """Sets the database to be used by this :class:`Document`.

        The database will also be used by subclasses of this class unless they :meth:`use` their own database.

        This method has to be invoked before the ODM class can be used.
        """
        assert db is not None
        cls.__db__ = db

    @classmethod
    def db(cls) -> AgnosticDatabase:
        """Returns the database that is currently associated with this document.

        If no such database exists this returns the database of the parent document (its superclass). If no
        :class:`Document` class had its :meth:`use` method called to set a db, an :class:`AttributeError` is raised.
        """
        if not hasattr(cls, "__db__"):
            raise AttributeError("Accessing database without using it first.")
        return cls.__db__

    @classmethod
    def collection(cls: Type["Document"]) -> AgnosticCollection:
        """Returns the collection for this :class:`Document`.

        The collection uses the ``codec_options``, ``read_preference``, ``write_concern`` and ``read_concern`` from the
        document's ```Mongo``` class.
        """
        meta = cls.__mongo__
        if cls.__collection__ is None or cls.__collection__.database is not cls.db():
            cls.__collection__ = cls.db().get_collection(
                meta.collection,
                codec_options=meta.codec_options,
                read_preference=meta.read_preference,
                write_concern=meta.write_concern,
                read_concern=meta.read_concern,
            )
        return cls.__collection__

    def document(
        self,
        *,
        include: Union["AbstractSetIntStr", "DictIntStrAny"] = None,
        exclude: Union["AbstractSetIntStr", "DictIntStrAny"] = None,
    ) -> "DictStrAny":
        """Converts this object into a dictionary suitable to be saved to MongoDB."""
        return self.dict(
            by_alias=True, include=include, exclude=exclude, exclude_defaults=True
        )

    @classmethod
    async def count(cls, db_filter: "Query" = None, **kwargs: Any) -> int:
        """Returns the number of documents in this class's collection.

        This method is filterable."""
        query = create_query(db_filter, **kwargs)
        return await cls.collection().count_documents(query)  # type: ignore

    @classmethod
    async def batch_insert(
        cls: Type["GenericDocument"], *objects: "GenericDocument"
    ) -> None:
        """Inserts multiple documents at once.

        It is preferred to use this method over multiple :meth:`insert` calls as the performance can be much better.
        """
        result: InsertManyResult = await cls.collection().insert_many(
            [obj.document() for obj in objects]
        )
        for obj, inserted_id in zip(objects, result.inserted_ids):
            obj.id = inserted_id

    @classmethod
    async def get(
        cls: Type["GenericDocument"], db_filter: "Query" = None, **kwargs: Any
    ) -> Optional["GenericDocument"]:
        """Returns a single document from the collection.

        This method is filterable.
        """
        query = create_query(db_filter, **kwargs)
        doc = await cls.collection().find_one(query)
        return cls(**doc) if doc else None

    @classmethod
    def find(
        cls: Type["GenericDocument"], db_filter: "Query" = None, **kwargs: Any
    ) -> AsyncIterator["GenericDocument"]:
        """Returns multiple documents from the collection.

        This method is filterable.
        """
        query = create_query(db_filter, **kwargs)

        async def context() -> AsyncIterator["GenericDocument"]:
            async for doc in cls.collection().find(query):
                yield cls(**doc)

        return context()

    all = find

    async def reload(self) -> None:
        """Reloads a document from the database.

        Use this method if a model might have changed in the database and you need to retrieve the current version. You
        do **not** need to call this after inserting a newly created object into the database.
        """
        updated = self.__class__(**await self.collection().find_one({"_id": self.id}))
        object.__setattr__(self, "__dict__", updated.__dict__)

    async def insert(self) -> None:
        """Inserts the object into the database.

        The object is inserted as a new object.
        """
        result = await self.collection().insert_one(self.document())
        self.id = result.inserted_id



import pytest

from motor_odm.query import create_query


def test_single_kwargs_query():
    query = create_query(username="test")
    assert query == {"username": "test"}


def test_multi_kwargs_query():
    query = create_query(username="test", password="abc")
    assert query == {"username": "test", "password": "abc"}


def test_filter():
    filter = {"something": "value", "more": {"subkey": "subvalue"}}
    query = create_query(filter)
    assert query == filter


@pytest.mark.parametrize("id", [123, "some id", 5.4])
def test_id(id):
    query = create_query(id)
    assert query == {"_id": id}


def test_id_and_kwargs():
    query = create_query(123, username="test", password="abc")
    assert query == {"_id": 123, "username": "test", "password": "abc"}


def test_filter_and_kwargs():
    query = create_query({"some": "field"}, username="test")
    assert query == {"some": "field", "username": "test"}



"""
This module contains functions for building MongoDB queries.
"""

from typing import TYPE_CHECKING, Any, Optional, Union

if TYPE_CHECKING:
    from pydantic.typing import DictStrAny

    Query = Optional[Union[DictStrAny, Any]]


def create_query(db_filter: "Query" = None, **kwargs: Any) -> "DictStrAny":
    """
    Creates a MongoDB query from the specified arguments. This helper can be invoked in three ways (which can also be
    combined):

    Create a filter from keyword arguments. The arguments are transformed into a :class:`dict` and returned verbatim.

    >>> create_query(username="john")
    {'username': 'john'}

    For more complex cases you can also supply your own filter as a :class:`dict`.

    >>> create_query({"username": "john"})
    {'username': 'john'}

    You can also combine both methods.

    >>> create_query({"username": "john"}, password="abc123")
    {'password': 'abc123', 'username': 'john'}

    One special case that is often required is looking up objects by their ID. You can do this by passing a value for
    the ID as the first positional argument. This cannot be combined with other advanced filters but can accept
    additional keyword arguments.

    >>> create_query("john", password="abc123")
    {'password': 'abc123', '_id': 'john'}

    :param db_filter: Either a dict that is included as a filter or any other type that is used as an ``_id`` query.
    :param kwargs: Filter arguments. Keyword arguments are preceeded by the ``db_filter`` parameter.
    :return: A :class:`dict` that can be used to filter a MongoDB collection.
    """
    query = kwargs
    if db_filter is None:
        pass
    elif not isinstance(db_filter, dict):
        query.update({"_id": db_filter})
    else:
        query.update(db_filter)
    return query



"""BSON encoders for common python types.

This module contains a collection of :class:`bson.codec_options.TypeEncoder` subclasses for common python types such as
sets. Note that these encoders are provided as a convenience but are not used automatically. If you want to use sets in
your documents you have to provide the appropriate ``codec_options`` to the MongoDB client, database, collection or
function.
"""

from typing import TYPE_CHECKING

from bson.codec_options import TypeEncoder

__all__ = ["SetEncoder", "FrozensetEncoder"]

if TYPE_CHECKING:
    from typing import TypeVar

    T = TypeVar("T")


class SetEncoder(TypeEncoder):
    """BSON support for python :class:`set`.

    This encoder encodes a :class:`set` in form of a :class:`list`. The list is not converted back into a set
    automatically but if you are using the :class:`Document <motor_odm.document.Document>` class this is done upon
    initialization of your model.
    """

    python_type = set
    transform_python = list


class FrozensetEncoder(TypeEncoder):
    """BSON support for python :class:`frozenset`.

    This encoder encodes a :class:`frozenset` in form of a :class:`list`. The list is not converted back into a set
    automatically but if you are using the :class:`Document <motor_odm.document.Document>` class this is done upon
    initialization of your model.
    """

    python_type = frozenset
    transform_python = list



from invoke import Context, task


@task
def test(c, coverage=False):
    c: Context
    args = ["pytest"]
    if coverage:
        args.extend(["--cov", "--cov-report=xml", "--cov-report=html"])
    c.run(" ".join(args))


@task
def doctest(c):
    c: Context
    c.run("pytest --doctest-modules --doctest-continue-on-failure motor_odm")


@task
def black(c, check=False):
    c: Context
    if check:
        c.run("black motor_odm --check")
    else:
        c.run("black motor_odm")


@task
def flake8(c):
    c: Context
    c.run("flake8")


@task
def mypy(c):
    c: Context
    c.run("mypy motor_odm")


@task
def isort(c, check=False):
    c: Context
    if check:
        c.run("isort -rc --check-only")
    else:
        c.run("isort -rc")


@task
def format(c):
    c: Context
    print("Formatting with Black...")
    black(c)
    print("Sorting Imports...")
    isort(c)


@task
def lint(c):
    c: Context
    print("Running Black...")
    black(c, check=True)
    print("Running flake8...")
    flake8(c)
    print("Running MyPy...")
    mypy(c)
    print("Running isort...")
    isort(c, check=True)


@task
def build_docs(c):
    c: Context
    c.run("cd docs && make html")


@task
def clean(c):
    c: Context
    files = [
        "build",
        "dist",
        "Motor_ODM.egg-info",
        "pip-wheel-metadata",
        "docs/build",
        "docs/reference",
        ".mypy_cache",
        ".pytest_cache",
        "htmlcov",
        ".coverage",
        "coverage.xml",
    ]
    args = ["rm", "-rf", *files]
    c.run(" ".join(args))



"""
This module contains patches for some frameworks to make Motor-ODM work as one would expect. Expect some more or less
ugly hacks here...

Note that all patches are applied automatically at **import time**.
"""

from typing import Any

from .document import Document
from .helpers import monkey_patch


def patch_fastapi() -> None:
    """
    Patches the `FastAPI <https://fastapi.tiangolo.com>`_ framework to support models based on Pydantic. By default
    FastAPI routes handle Pydantic models specially. This patch removes the special case for subclasses of
    :class:`Document <motor_odm.document.Document>`.
    """
    import fastapi  # noqa

    @monkey_patch(fastapi.routing)
    def _prepare_response_content(res: Any, *args: Any, **kwargs: Any) -> Any:
        if isinstance(res, Document):
            return res
        return _prepare_response_content.original(res, **kwargs)  # type: ignore


try:
    patch_fastapi()
except ImportError:
    pass



import os

from setuptools import find_packages, setup


def long_description():
    this_directory = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
        return f.read()


def list_requirements(file):
    with open(file) as f:
        return [line.strip() for line in f.readlines() if "." not in line]


setup(
    name="Motor-ODM",
    url="https://github.com/Codello/Motor-ODM/",
    license="MIT",
    author="Kim Wittenburg",
    author_email="codello@wittenburg.kim",
    description="A MongoDB ODM based on Motor and Pydantic.",
    long_description=long_description(),
    long_description_content_type="text/markdown",
    keywords="MongoDB AsyncIO ODM Pydantic",
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    platforms="any",
    use_scm_version={
        "local_scheme": "no-local-version",
    },  # Support Legacy Installations
    install_requires=["PyMongo", "Motor", "Pydantic"],
    extras_require={
        "docs": list_requirements("docs/requirements.txt"),
        "dev": [
            "pytest",
            "pytest-cov",
            "pytest-asyncio",
            "flake8",
            "isort",
            "mypy",
            "black",
            "invoke",
        ],
        "typing": ["fastapi"],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Database",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import pytest

from motor_odm import Document


def test_create_valid_document():
    class Test(Document):
        class Mongo:
            collection = "test"

        field: str

    Test(field="")


def test_create_document_without_mongo_class():
    with pytest.raises(TypeError):

        class Test(Document):
            field: str
